/*
 * chardev.c: 创建一个只读 char 设备，显示从 dev 文件读取的次数
 */

#include <linux/atomic.h>
#include <linux/cdev.h>
#include <linux/delay.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/kernel.h> /* 引入 sprintf() */
#include <linux/module.h>
#include <linux/printk.h>
#include <linux/types.h>
#include <linux/uaccess.h> /* 引入 get_user 和 put_user */
#include <linux/version.h>

#include <asm/errno.h>

/*  原型(Prototypes) - 这通常会放在一个 .h 文件中 */
static int device_open(struct inode *, struct file *);
static int device_release(struct inode *, struct file *);
static ssize_t device_read(struct file *, char __user *, size_t, loff_t *);
static ssize_t device_write(struct file *, const char __user *, size_t,
                            loff_t *);

#define SUCCESS 0
#define DEVICE_NAME "chardev" /* 在/proc/devices中显示的Dev名称 */
#define BUF_LEN 80 /* 来自驱动消息的最大长度 */

/* 全局变量声明为静态的, 文件中的全局变量也是如此 */

static int major; /* 分配给设备驱动程序的主编号 */

enum {
    CDEV_NOT_USED = 0,
    CDEV_EXCLUSIVE_OPEN = 1,
};

/* 设备是否打开？用于防止多次访问设备 */
static atomic_t already_open = ATOMIC_INIT(CDEV_NOT_USED);

static char msg[BUF_LEN + 1]; /* 当被请求时，设备将给出的信息 */

static struct class *cls;

static struct file_operations chardev_fops = {
    .read = device_read,
    .write = device_write,
    .open = device_open,
    .release = device_release,
};

static int __init chardev_init(void)
{
    major = register_chrdev(0, DEVICE_NAME, &chardev_fops);

    if (major < 0) {
        pr_alert("注册字符设备失败 %d\n", major);
        return major;
    }

    pr_info("分配的主编号 %d.\n", major);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 4, 0)
    cls = class_create(DEVICE_NAME);
#else
    cls = class_create(THIS_MODULE, DEVICE_NAME);
#endif
    device_create(cls, NULL, MKDEV(major, 0), NULL, DEVICE_NAME);

    pr_info("设备创建于 /dev/%s\n", DEVICE_NAME);

    return SUCCESS;
}

static void __exit chardev_exit(void)
{
    device_destroy(cls, MKDEV(major, 0));
    class_destroy(cls);

    /* 注销设备 */
    unregister_chrdev(major, DEVICE_NAME);
}

/* 方法 */

/* 当进程尝试打开设备文件时调用，如
 * "sudo cat /dev/chardev"
 */
static int device_open(struct inode *inode, struct file *file)
{
    static int counter = 0;

    if (atomic_cmpxchg(&already_open, CDEV_NOT_USED, CDEV_EXCLUSIVE_OPEN))
        return -EBUSY;

    sprintf(msg, "我已经跟你说了 %d 次 Hello world!\n", counter++);
    try_module_get(THIS_MODULE);

    return SUCCESS;
}

/* 当进程关闭设备文件时调用 */
static int device_release(struct inode *inode, struct file *file)
{
    /* 准备好迎接下一个触发 */
    atomic_set(&already_open, CDEV_NOT_USED);

    /* 减少使用次数，否则一旦打开文件，就永远无法删除模块 */
    module_put(THIS_MODULE);

    return SUCCESS;
}

/* 当一个已打开 dev 文件的进程试图读取该文件时调用 */
static ssize_t device_read(struct file *filp, /* 请参见 include/linux/fs.h   */
                           char __user *buffer, /* 待填充数据的缓冲区 */
                           size_t length, /* 缓冲区长度 */
                           loff_t *offset)
{
    /* 实际写入缓冲区的字节数 */
    int bytes_read = 0;
    const char *msg_ptr = msg;

    if (!*(msg_ptr + *offset)) { /* 已到达消息的末尾 */
        *offset = 0; /* 重置偏移量 */
        return 0; /* 表示文件结束 */
    }

    msg_ptr += *offset;

    /* 实际将数据放入缓冲区 */
    while (length && *msg_ptr) {
        /* 缓冲区在用户数据段，而不是内核段所以直接赋值 "*" 是无效的
         * 我们必须使用 put_user 函数将数据从内核数据段复制到用户数据段
         */
        put_user(*(msg_ptr++), buffer++);
        length--;
        bytes_read++;
    }

    *offset += bytes_read;

    /* 大多数读取函数返回放入缓冲区的字节数 */
    return bytes_read;
}

/* 当进程写入dev文件时调用: echo "hi" > /dev/hello */
static ssize_t device_write(struct file *filp, const char __user *buff,
                            size_t len, loff_t *off)
{
    pr_alert("对不起，不支持此操作.\n");
    return -EINVAL;
}

module_init(chardev_init);
module_exit(chardev_exit);

MODULE_LICENSE("GPL");

