/*
 * syscall-steal.c
 * 系统调用 "监视" 示例。
 * 通过更改 cr0 寄存器中的第 16 位，在处理器级别禁用页面保护(可能与 Intel 有关)
 */

#include <linux/delay.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h> 	/* 包含一些参数 */
#include <linux/unistd.h> 		/* 系统调用列表 */
#include <linux/cred.h> 		/* 引入 current_uid() */
#include <linux/uidgid.h> 		/* 引入 __kuid_val() */
#include <linux/version.h>

/* 根据本程序的需求我们需要了解当前用户是谁 */
#include <linux/sched.h>
#include <linux/uaccess.h>

/* 访问“sys_call_table”的方式随着内核内部的变化而变化。
 * 内核版本 <= v5.4 		: 手动符号查找
 * 内核版本 == v5.5 或 v5.6 : 使用 kallsyms_lookup_name()
 * 内核版本 >= v5.7         : Kprobes 或特定的内核模块参数
 */

/* 在 Linux v5.11 及以上版本中，内核调用 ksys_close() 的方式被移除 */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0))

#if LINUX_VERSION_CODE <= KERNEL_VERSION(5, 4, 0)
#define HAVE_KSYS_CLOSE 1
#include <linux/syscalls.h> /* 引入 ksys_close() */
#else
#include <linux/kallsyms.h> /* 引入 kallsyms_lookup_name */
#endif

#else

#if defined(CONFIG_KPROBES)
#define HAVE_KPROBES 1
#if defined(CONFIG_X86_64)
/* 如果你尝试使用系统调用表来拦截系统调用但未成功，你可以尝试使用 Kprobes 来拦截系统调用。
 * 将 USE_KPROBES_PRE_HANDLER_BEFORE_SYSCALL 设置为 1 以注册一个预处理程序在系统调用之前。
 */
#define USE_KPROBES_PRE_HANDLER_BEFORE_SYSCALL 0
#endif
#include <linux/kprobes.h>
#else
#define HAVE_PARAM 1
#include <linux/kallsyms.h> /* 引入 sprint_symbol */
/* sys_call_table 的地址可以通过查看 "/boot/System.map" 或 "/proc/kallsyms" 获得。
 * 当内核版本为 v5.7 及以上时，如果没有 CONFIG_KPROBES，你可以输入参数或者模块会查找所有内存。
 */
static unsigned long sym = 0;
module_param(sym, ulong, 0644);
#endif /* CONFIG_KPROBES */

#endif /* 版本 < v5.7 */

/* 我们想要监视的 UID - 将从命令行填充。 */
static uid_t uid = -1;
module_param(uid, int, 0644);

#if USE_KPROBES_PRE_HANDLER_BEFORE_SYSCALL

/* syscall_sym 是要监视的系统调用的符号名称。默认值为"__x64_sys_openat"
 * 可以通过模块参数进行更改。你可以在 /proc/kallsyms 中查找系统调用的符号名称。
 */
static char *syscall_sym = "__x64_sys_openat";
module_param(syscall_sym, charp, 0644);

static int sys_call_kprobe_pre_handler(struct kprobe *p, struct pt_regs *regs)
{
    if (__kuid_val(current_uid()) != uid) {
        return 0;
    }

    pr_info("%s 被 %d 调用\n", syscall_sym, uid);
    return 0;
}

static struct kprobe syscall_kprobe = {
    .symbol_name = "__x64_sys_openat",
    .pre_handler = sys_call_kprobe_pre_handler,
};
#else

static unsigned long **sys_call_table_stolen;

/* 指向原始系统调用的指针。我们保留这个指针，而不是直接调用原始函数 (sys_openat)，是因为可能有其他模块在我们之前已经替换了系统调用。
 * 注意，这不是 100% 安全的，因为如果其他模块在我们之前替换了 sys_openat，那么当我们插入时，我们将调用那个模块中的函数 -
 * 而且那个模块可能在我们之前已经被移除。
 *
 * 另一个原因是我们无法获取 sys_openat。它是一个静态变量，因此没有被导出。
 */
#ifdef CONFIG_ARCH_HAS_SYSCALL_WRAPPER
static asmlinkage long (*original_call)(const struct pt_regs *);
#else
static asmlinkage long (*original_call)(int, const char __user *, int, umode_t);
#endif

/* 我们将用来替换 sys_openat (即调用 open 系统调用时调用的函数)的函数
 * 为了找到准确的原型，包含参数的数量和类型，我们首先找到原始函数(它位于 fs/open.c 文件中)
 *
 * 理论上，这意味着我们绑定于当前版本的内核。实际上，系统调用几乎不会改变
 * (因为系统调用是内核和进程之间的接口，如果系统调用发生变化，会引起混乱并需要重新编译程序)
 */
#ifdef CONFIG_ARCH_HAS_SYSCALL_WRAPPER
static asmlinkage long our_sys_openat(const struct pt_regs *regs)
#else
static asmlinkage long our_sys_openat(int dfd, const char __user *filename,
                                      int flags, umode_t mode)
#endif
{
    int i = 0;
    char ch;

    if (__kuid_val(current_uid()) != uid)
        goto orig_call;

    /* 报告相关文件 */
    pr_info("Opened file by %d: ", uid);
    do {
#ifdef CONFIG_ARCH_HAS_SYSCALL_WRAPPER
        get_user(ch, (char __user *)regs->si + i);
#else
        get_user(ch, (char __user *)filename + i);
#endif
        i++;
        pr_info("%c", ch);
    } while (ch != 0);
    pr_info("\n");

orig_call:
    /* 调用原始的sys_openat-否则，我们将失去打开文件的能力 */
#ifdef CONFIG_ARCH_HAS_SYSCALL_WRAPPER
    return original_call(regs);
#else
    return original_call(dfd, filename, flags, mode);
#endif
}

static unsigned long **acquire_sys_call_table(void)
{
#ifdef HAVE_KSYS_CLOSE
    unsigned long int offset = PAGE_OFFSET;
    unsigned long **sct;

    while (offset < ULLONG_MAX) {
        sct = (unsigned long **)offset;

        if (sct[__NR_close] == (unsigned long *)ksys_close)
            return sct;

        offset += sizeof(void *);
    }

    return NULL;
#endif

#ifdef HAVE_PARAM
    const char sct_name[15] = "sys_call_table";
    char symbol[40] = { 0 };

    if (sym == 0) {
        pr_alert("对于 Linux v5.7 以上版本，Kprobes 是获取符号的首选方法\n");
        pr_info("如果没有 Kprobes，则必须指定 sys_call_table 符号的地址\n");
        pr_info("通过 /boot/System.map 或 /proc/kallsyms 将包含所有符号地址的符号转换为sym参数\n");
        return NULL;
    }
    sprint_symbol(symbol, sym);
    if (!strncmp(sct_name, symbol, sizeof(sct_name) - 1))
        return (unsigned long **)sym;

    return NULL;
#endif

#ifdef HAVE_KPROBES
    unsigned long (*kallsyms_lookup_name)(const char *name);
    struct kprobe kp = {
        .symbol_name = "kallsyms_lookup_name",
    };

    if (register_kprobe(&kp) < 0)
        return NULL;
    kallsyms_lookup_name = (unsigned long (*)(const char *name))kp.addr;
    unregister_kprobe(&kp);
#endif

    return (unsigned long **)kallsyms_lookup_name("sys_call_table");
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 3, 0)
static inline void __write_cr0(unsigned long cr0)
{
    asm volatile("mov %0,%%cr0" : "+r"(cr0) : : "memory");
}
#else
#define __write_cr0 write_cr0
#endif

static void enable_write_protection(void)
{
    unsigned long cr0 = read_cr0();
    set_bit(16, &cr0);
    __write_cr0(cr0);
}

static void disable_write_protection(void)
{
    unsigned long cr0 = read_cr0();
    clear_bit(16, &cr0);
    __write_cr0(cr0);
}
#endif

static int __init syscall_steal_start(void)
{
#if USE_KPROBES_PRE_HANDLER_BEFORE_SYSCALL

    int err;
    /* use symbol name from the module parameter */
    syscall_kprobe.symbol_name = syscall_sym;
    err = register_kprobe(&syscall_kprobe);
    if (err) {
        pr_err("register_kprobe() 在 %s 失败: %d\n", syscall_sym, err);
        pr_err("请检查 “syscall_sym ”参数中的符号名称\n");
        return err;
    }

#else
    if (!(sys_call_table_stolen = acquire_sys_call_table()))
        return -1;

    disable_write_protection();

    /* 跟踪原始的打开函数 */
    original_call = (void *)sys_call_table_stolen[__NR_openat];

    /* 请改用我们的openat函数 */
    sys_call_table_stolen[__NR_openat] = (unsigned long *)our_sys_openat;

    enable_write_protection();

#endif

    pr_info("监视 UID:%d\n", uid);
    return 0;
}

static void __exit syscall_steal_end(void)
{
#if USE_KPROBES_PRE_HANDLER_BEFORE_SYSCALL
    unregister_kprobe(&syscall_kprobe);
#else
    if (!sys_call_table_stolen)
        return;

    /* 将系统调用恢复正常 */
    if (sys_call_table_stolen[__NR_openat] != (unsigned long *)our_sys_openat) {
        pr_alert("发现系统调用被其他人修改");
        pr_alert("系统可能处于不稳定状态");
    }

    disable_write_protection();
    sys_call_table_stolen[__NR_openat] = (unsigned long *)original_call;
    enable_write_protection();
#endif

    msleep(2000);
}

module_init(syscall_steal_start);
module_exit(syscall_steal_end);

MODULE_LICENSE("GPL");
