/*
 * intrpt.c - 处理带中断的 GPIO
 *
 * 基于 Stefan Wendler (devnull@kaltpost.de) 提供的 RPi 示例：
 *   https://github.com/wendlers/rpi-kmod-samples
 *
 * 按下一个按钮点亮 LED，按下另一个按钮熄灭 LED。
 */

#include <linux/gpio.h>
#include <linux/interrupt.h>
#include <linux/kernel.h> /* 引入 ARRAY_SIZE() */
#include <linux/module.h>
#include <linux/printk.h>

static int button_irqs[] = { -1, -1 };

/* 定义 LED 的 GPIO。
 * TODO: 根据你的板子更改 GPIO 编号。
 */
static struct gpio leds[] = { { 1, GPIOF_OUT_INIT_LOW, "LED 1" } };

/* 定义按钮的 GPIO。
 * TODO: 根据你的板子更改 GPIO 编号。
 */
static struct gpio buttons[] = { { 2, GPIOF_IN, "LED 1 打开按钮" },
                                 { 3, GPIOF_IN, "LED 1 关闭按钮" } };

/* 按钮按下时触发的中断函数。 */
static irqreturn_t button_isr(int irq, void *data)
{
    /* 第一个按钮 */
    if (irq == button_irqs[0] && !gpio_get_value(leds[0].gpio))
        gpio_set_value(leds[0].gpio, 1);  /* 点亮 LED */
    /* 第二个按钮 */
    else if (irq == button_irqs[1] && gpio_get_value(leds[0].gpio))
        gpio_set_value(leds[0].gpio, 0);  /* 熄灭 LED */

    return IRQ_HANDLED;
}

static int __init intrpt_init(void)
{
    int ret = 0;

    pr_info("%s\n", __func__);

    /* 注册 LED GPIO */
    ret = gpio_request_array(leds, ARRAY_SIZE(leds));

    if (ret) {
        pr_err("无法请求 LED 的 GPIO: %d\n", ret);
        return ret;
    }

    /* 注册按钮 GPIO */
    ret = gpio_request_array(buttons, ARRAY_SIZE(buttons));

    if (ret) {
        pr_err("无法请求按钮的 GPIO: %d\n", ret);
        goto fail1;
    }

    pr_info("当前按钮1的值: %d\n", gpio_get_value(buttons[0].gpio));

    ret = gpio_to_irq(buttons[0].gpio);

    if (ret < 0) {
        pr_err("无法请求 IRQ: %d\n", ret);
        goto fail2;
    }

    button_irqs[0] = ret;

    pr_info("成功请求 BUTTON1 的 IRQ # %d\n", button_irqs[0]);
	
	/* 请求按钮1的中断
	 * IRQF_TRIGGER_RISING 表示在信号电平从低到高变化时触发中断
	 * IRQF_TRIGGER_FALLING 表示在信号电平从高到低变化时触发中断
	 * 这种组合使得中断处理函数在检测到按钮的按下和释放时都会被调用。 
	 * */
    ret = request_irq(button_irqs[0], button_isr,
                      IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
                      "gpio模块#按钮1", NULL);

    if (ret) {
        pr_err("无法请求 IRQ: %d\n", ret);
        goto fail2;
    }

    ret = gpio_to_irq(buttons[1].gpio);

    if (ret < 0) {
        pr_err("无法请求 IRQ: %d\n", ret);
        goto fail2;
    }

    button_irqs[1] = ret;

    pr_info("成功请求 BUTTON2 的 IRQ # %d\n", button_irqs[1]);
	
	/* 请求按钮2的中断, 标志位同上 */
    ret = request_irq(button_irqs[1], button_isr,
                      IRQF_TRIGGER_RISING | IRQF_TRIGGER_FALLING,
                      "gpio模块#按钮2", NULL);

    if (ret) {
        pr_err("无法请求 IRQ: %d\n", ret);
        goto fail3;
    }

    return 0;

/* 清理已经设置的部分 */
fail3:
    free_irq(button_irqs[0], NULL);					/* 释放按钮1的中断请求 */

fail2:
    gpio_free_array(buttons, ARRAY_SIZE(buttons));	/* 释放按钮的GPIO */

fail1:
    gpio_free_array(leds, ARRAY_SIZE(leds));		/* 释放LED的GPIO */

    return ret;
}

static void __exit intrpt_exit(void)
{
    int i;

    pr_info("%s\n", __func__);

    /* 释放 IRQ */
    free_irq(button_irqs[0], NULL);
    free_irq(button_irqs[1], NULL);

    /* 关闭所有 LED */
    for (i = 0; i < ARRAY_SIZE(leds); i++)
        gpio_set_value(leds[i].gpio, 0);

    /* 注销 GPIO */
    gpio_free_array(leds, ARRAY_SIZE(leds));
    gpio_free_array(buttons, ARRAY_SIZE(buttons));
}

module_init(intrpt_init);
module_exit(intrpt_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("处理一些 GPIO 中断");

